# DE_analysis_pipeline.R
# author: Francesco Garassino

# Automates all steps required to perform and visualize the per-species
# differential expression (DE) analysis of gene expression

# DESeq manual accessed at https://bioconductor.org/packages/release/bioc/manuals/DESeq2/man/DESeq2.pdf

library(tidyverse)
library(DESeq2)

data_path <- "C:/Users/garas001/OneDrive - Wageningen University & Research/PhD/_Year 4/Chapter 3 - RNA-Seq Hirschfeldia/1_Alignment_quantification_nuclear_organelles_140bp/gene count matrices/"
output_path <- "C:/Users/garas001/OneDrive - Wageningen University & Research/PhD/_Year 4/Chapter 3 - RNA-Seq Hirschfeldia/2_DE_analysis/R outputs_04_05/"
dir.create(output_path, showWarnings = FALSE)


count_matrices <- list.files(path = data_path)
count_matrices <- count_matrices[c(1,3,2,4)] # reorder
matrices_paths <- paste(data_path, count_matrices, sep = "")

stats_table <- tibble(
  species = character(),
  total_genes = numeric(),
  expressed_genes = numeric(),
  genes_up = numeric(),
  pct_genes_up_tot = numeric(),
  pct_genes_up_exp = numeric(),
  genes_down = numeric(),
  pct_genes_down_tot = numeric(),
  pct_genes_down_exp = numeric()
  )

pca_list <- list()
ma_list <- list()

for (path in matrices_paths) {
  file_name = tail(unlist(strsplit(path, "/")), 1)
  species = head(unlist(strsplit(file_name, "_longest")), 1)
  species_hr = sub("_", ". ", species)
  
  print(paste("Processing", species_hr, "count matrix..."))
  
  exp_data = read.csv(path, row.names = "gene_id", header = TRUE)  %>% 
    rename_with(~ gsub("X", "", .x)) # remove that weird "X" at the beginning of col. names

  lib_sizes <- round(colSums(exp_data)/10^6, 2)
  avg_lib_size <- mean(lib_sizes)
  sd_lib_size <- sd(lib_sizes)
  
  ## DE Analysis
  
  # grouping samples
  samples_treatments <- tibble(sample = colnames(exp_data)) %>% 
    mutate(treatment = factor(case_when(
      substr(sample,1,1) == 1 ~ "LL",
      substr(sample,1,1) == 2 ~ "HL"))) %>% 
    column_to_rownames(var = "sample")
 
  # Creating the DESeqDataSet (dds) object from count matrix
  dds <- DESeqDataSetFromMatrix(exp_data, samples_treatments, ~treatment)
  dds$treatment <- relevel(dds$treatment, "LL") # set LL as reference condition
  
  # The estimation of per-sample normalisation factors is the first step of the DE analysis
  # It basically measures how "big" or "small" sample libraries are with respect to each other, 
  # and will allow us to perform a correction for different library sizes and RNA composition.
  # Standard method is "ratio". Further info at DESeq manual page 22.
  
  dds <- estimateSizeFactors(dds, type = "ratio") 
  sizeFactors(dds)
  
  # Count data transformation - handy for visualisation purposes
  # The rlog (regularized logarithm) function transforms the count data to the log2 scale in a way which minimizes 
  # differences between samples for rows with small counts, and which normalizes with respect to library size.
  # Further info at DESeq manual page 53.
  
  counts_rlog_deseq <- rlog(dds, blind = FALSE) 
  # blind dispersion estimation is not the appropriate choice if one expects that many or the majority of genes (rows) will 
  # have large differences in counts which are explainable by the experimental design
  
  ## Visualization of samples groupings with PCA
  # For more info on plotPCA, see DESeq2 manual page 44
  
  colors_treatments <- c('HL' = '#D41159', 'LL' = '#1A85FF')  
  # https://davidmathlogic.com/colorblind/#%231A85FF-%23D41159
  
  p_pca <- DESeq2::plotPCA(counts_rlog_deseq, 
                           ntop = 500, 
                           intgroup = "treatment", 
                           returnData = FALSE) +
    geom_hline(yintercept = 0, linetype = 2) +
    geom_vline(xintercept = 0, linetype = 2) +
    scale_color_manual(values = colors_treatments) +
    ggrepel::geom_text_repel(aes(label = name), 
                             color = "black", 
                             force_pull = 0.5, 
                             size = 5) +
    scale_y_continuous(limits = c(-22,22), breaks=seq(-20,20,5)) +
    scale_x_continuous(limits = c(-40,40), breaks=seq(-40,40,10)) +
    theme_grey(base_size = 16) +
    theme(legend.position = "none")
  
  pca_list <- append(pca_list, list(p_pca))
  
  ggsave(filename = paste(species, "_pca_plot.jpg", sep = ""), 
         plot = p_pca, 
         device = "jpg", 
         path = output_path,
         scale = 3,
         width = 823,
         height = 550,
         units = "px")
  
  ## Visualization of samples groupings with Hierarchical clustering
  cor_dist <- factoextra::get_dist(t(assay(counts_rlog_deseq)), method = "pearson") # note the distance metric based on correlation
  hc <- hclust(cor_dist)
  hcd <- as.dendrogram(hc)
  
  jpeg(filename = paste(output_path, species, "_hc_plot_corr.jpg", sep = ""),
       width = 850,
       height = 550,
       units = "px", 
       pointsize = 24,
       quality = 100)
  plot(hcd, type = "rectangle", ylab = "Height")
  dev.off()
  
  ## DE analysis continued
  
  # EstimateDispersions obtains dispersion estimates for Negative Binomial distributed data
  # Additional info at DESeq2 manual page 17
  dds <- estimateDispersions(dds)
  
  plotDispEsts(dds, main = species)
  
  # Now, we test for significance of coefficients in a Negative Binomial GLM, using previously calculated 
  # sizeFactors and dispersion estimates. 
  # Additional info at DESeq2 manual page 35
  dds <- nbinomWaldTest(dds)
  # for results(), see DESeq manual page 48
  res <- results(dds, alpha = 0.05)
  
  ## visualization with MA plot
  p_MA <- ggpubr::ggmaplot(res, 
                    #main = paste(species_hr, "- expression under HL relative to LL"),
                    fdr = 0.05, 
                    fc = 0, 
                    size = 2,
                    #alpha = 0.5,
                    palette = c("#D41159", "#1A85FF", "darkgray"),
                    genenames = rownames(res),
                    legend = "top", 
                    top = 0, # no gene labels
                    select.top.method = "fc",
                    font.label = c("bold", 11), label.rectangle = TRUE,
                    font.legend = "bold",
                    font.main = "bold",
                    #ylim = c(-8, 12),
                    #xlim = c(1.8, 20),
                    ggtheme = ggplot2::theme_minimal(base_size = 16)
                   ) +
    scale_y_continuous(limits = c(-25,25), breaks=seq(-25,25,5)) +
    scale_x_continuous(limits = c(0,20), breaks=seq(0,20,2))
  
  ma_list <- append(ma_list, list(p_MA))
  
  ggsave(filename = paste(species, "_MA_plot.jpg", sep = ""), 
         plot = p_MA, 
         device = "jpg", 
         path = output_path,
         scale = 3,
         width = 823,
         height = 550,
         units = "px")
  
  ## Gather stats and export results
  res_df <- as.data.frame(res) 
  total_genes = nrow(res_df)
  expressed_genes = nrow(filter(res_df, !is.na(padj)))
  genes_up = res_df %>% filter(log2FoldChange > 0 & padj < 0.05) %>% nrow()
  pct_genes_up_tot = genes_up/total_genes * 100
  pct_genes_up_exp = genes_up/expressed_genes * 100
  genes_down = res_df %>% filter(log2FoldChange < 0 & padj < 0.05) %>% nrow()
  pct_genes_down_tot = genes_down/total_genes * 100
  pct_genes_down_exp = genes_down/expressed_genes * 100
  
  stats_table <- stats_table %>% 
    bind_rows(., tibble(species, total_genes, expressed_genes, 
                        genes_up, pct_genes_up_tot, pct_genes_up_exp,
                        genes_down, pct_genes_down_tot, pct_genes_down_exp))
  
  # write reults to table
  write.csv(res_df,
            file = paste(output_path, species, "_DE_results.csv", sep = ""),
            row.names = TRUE)
  
  print(paste("Finished processing", species_hr, "count matrix..."))
}

# export stats_table
 write_csv(stats_table,
          file = paste(output_path, "All_species_genes_stats.csv", sep = ""))

# figure with PCA plots
pca_grid <- cowplot::plot_grid(plotlist = pca_list,
                   label_fontfamily = "sans", # sans = Arial (Windows)
                   label_size = 20,
                   labels = "auto")

ggsave("PCA_plots_facet.png",
       plot = pca_grid,
       device = "png",
       path = output_path,
       width = 9.4,
       height = 7.3,
       units = "in")

# figure with MA plots
ma_grid <- cowplot::plot_grid(plotlist = ma_list,
                   label_fontfamily = "sans",
                   label_size = 20,
                   labels = "auto")

ggsave("MA_plots_facet.png",
       plot = ma_grid,
       device = "png",
       path = output_path,
       width = 12.5,
       height = 8.9,
       units = "in")
