library(tidyverse)

out_path = "outputs/"

# read in the orthology/frequency table resulting from the pan-genome
ort_freq_table <- read_csv("data/65_orthology_frequency_table_renamed.csv", 
                           show_col_types = FALSE) %>% 
  janitor::clean_names() %>% 
  mutate(gene_id_a_thaliana = gsub("\\.Araport11.447", "", gene_id_a_thaliana)) # fix the A. thaliana gene names

# read in the count matrix and DE table for A. thaliana, rename columns, make gene ID shorter and more relevant
count_matrix_a_thaliana <- read_csv("data/A_thaliana_longest_isoforms_filter_gene_count_matrix.csv",
                            show_col_types = FALSE) %>% 
  rename(LL1 = `1a4`,
         LL2 = `1b3`,
         LL3 = `1c4`,
         LL4 = `1d2`,
         LL5 = `1f3`,
         HL1 = `2a4`,
         HL2 = `2b3`,
         HL3 = `2c2`,
         HL4 = `2d3`,
         HL5 = `2f2`) %>% 
  mutate(gene_id = gsub("\\..*", "", gene_id)) %>% 
  rename_with(~paste0("at_", .), -gene_id)

de_table_a_thaliana <- read_csv("data/A_thaliana_longest_isoforms_filter_DE_results.csv",
                        show_col_types = FALSE) %>% 
  rename(gene_id = `...1`) %>% 
  mutate(gene_id = gsub("\\..*", "", gene_id)) %>% 
  rename_with(~paste0("at_", .), -gene_id)

# read in the count matrix and DE table for B. rapa, rename columns, make gene ID shorter and more relevant
count_matrix_b_rapa <- read_csv("data/B_rapa_longest_isoforms_filter_gene_count_matrix.csv",
                            show_col_types = FALSE) %>% 
  rename(LL1 = `1a2`,
         LL2 = `1c3`,
         LL3 = `1d5`,
         LL4 = `1e2`,
         LL5 = `1e4`,
         HL1 = `2b2`,
         HL2 = `2c3`,
         HL3 = `2d4`,
         HL4 = `2e2`,
         HL5 = `2f4`) %>% 
  rename_with(~paste0("br_", .), -gene_id)

de_table_b_rapa <- read_csv("data/B_rapa_longest_isoforms_filter_DE_results.csv",
                        show_col_types = FALSE) %>% 
  rename(gene_id = `...1`) %>% 
  rename_with(~paste0("br_", .), -gene_id)

# read in the count matrix and DE table for B. nigra, rename columns, make gene ID shorter and more relevant
count_matrix_b_nigra <- read_csv("data/B_nigra_longest_isoforms_filter_gene_count_matrix.csv",
                                show_col_types = FALSE) %>% 
  rename(LL1 = `1a3`,
         LL2 = `1b4`,
         LL3 = `1c2`,
         LL4 = `1e5`,
         LL5 = `1f2`,
         HL1 = `2a3`,
         HL2 = `2b4`,
         HL3 = `2c5`,
         HL4 = `2d2`,
         HL5 = `2e3`) %>% 
  rename_with(~paste0("bn_", .), -gene_id)

de_table_b_nigra <- read_csv("data/B_nigra_longest_isoforms_filter_DE_results.csv",
                            show_col_types = FALSE) %>% 
  rename(gene_id = `...1`) %>%
  rename_with(~paste0("bn_", .), -gene_id)

# read in the count matrix and DE table for H. incana, rename columns, make gene ID shorter and more relevant
count_matrix_h_incana <- read_csv("data/H_incana_longest_isoforms_filter_gene_count_matrix.csv",
                                 show_col_types = FALSE) %>% 
  rename(LL1 = `1b2`,
         LL2 = `1c5`,
         LL3 = `1d4`,
         LL4 = `1e3`,
         LL5 = `1f4`,
         HL1 = `2a2`,
         HL2 = `2c4`,
         HL3 = `2d5`,
         HL4 = `2e4`,
         HL5 = `2f3`) %>% 
  rename_with(~paste0("hi_", .), -gene_id)

de_table_h_incana <- read_csv("data/H_incana_longest_isoforms_filter_DE_results.csv",
                             show_col_types = FALSE) %>% 
  rename(gene_id = `...1`) %>%
  rename_with(~paste0("hi_", .), -gene_id)

# 1 - extract core groups (at least one copy, all species)
core_groups_all <- ort_freq_table %>%
  filter(if_all(starts_with("copy_nr"), ~ . > 0))

write_csv(core_groups_all,
          file = paste0(out_path, "core_groups_all_species.csv"))

# 2 - extract single-copy core groups (exactly one copy, all species)
sc_core_groups_all <- core_groups_all %>% 
  filter(if_all(starts_with("copy_nr"), ~ . == 1))

write_csv(sc_core_groups_all,
          file = paste0(out_path, "single_copy_core_groups_all_species.csv"))

# 3 - extract "core" groups for the four species (at least one copy)
core_groups_4species <- ort_freq_table %>% 
  filter(if_all(c('copy_nr_a_thaliana', 'copy_nr_b_rapa', 'copy_nr_b_nigra', 'copy_nr_h_incana'), ~ . > 0))

write_csv(core_groups_4species,
          file = paste0(out_path, "core_groups_4_species.csv"))

# 4 - extract single-copy "core" groups for the four species (exactly one copy)
sc_core_groups_4species <- core_groups_4species %>% 
  filter(if_all(c('copy_nr_a_thaliana', 'copy_nr_b_rapa', 'copy_nr_b_nigra', 'copy_nr_h_incana'), ~ . == 1))

write_csv(sc_core_groups_4species,
          file = paste0(out_path, "single_copy_core_groups_4_species.csv"))

# merge with count tables and DE tables!
# species_list = c("a_thaliana", "b_nigra", "b_rapa", "h_incana")
# 
# Sys.time()
# max_rows <- nrow(sc_core_groups_4species)
# #max_rows = 50
# for (nrow in 1:max_rows){
#   
#   # print a first message & do some cleanup, just in case...
#   if(nrow == 1){
#     cat("Started processing\n")
#     cat(paste0("Orthology table has ", max_rows, " rows\n\n"))
#     
#     if(exists("sc_core_4species_merged_table")) {
#       rm(sc_core_4species_merged_table)
#     }
#   }
#   
#   # print status messages
#   if(nrow%%100 ==0){
#     cat(paste0("\t processed ", nrow, " rows\n"))
#   }
#   
#   pct = (nrow / max_rows)*100
#   if(pct %in% c(10,20,30,40,50,60,70,80,90)){
#     cat(paste0("\t\t", pct, "% of processing done...\n"))
#   }
#   
#   # extract current row
#   row <- slice(sc_core_groups_4species, nrow)
#   
#   # iterate over species to retrieve count data
#   for(spc in species_list){
#     # here all gene IDs will match with the count table
#     row <- bind_cols(row,
#                      eval(parse(text = paste0("count_matrix_", spc))) %>% 
#                        filter(grepl(row %>% 
#                                       select(paste0("gene_id_", spc)) %>% 
#                                       sub("\\..*", "", .), 
#                                     gene_id)) %>% 
#                        select(!gene_id)
#                      )
#   }
#   
#   # iterate over species to retrieve DE values
#   for(spc in species_list){
#     # check if the gene ID is present in the DE table (of not, the gene was filtered out)
#     test_presence <- grepl(
#       row %>% select(paste0("gene_id_", spc)) %>% sub("\\..*", "", .),
#         eval(parse(text = paste0("de_table_", spc))) %>% select(gene_id))
#     
#     if(!test_presence){
#       # if gene is not present in DE table, make sure NAs are inserted
#       row <- bind_cols(row,
#                        colnames(eval(parse(text = paste0("de_table_", spc)))) %>%
#                          purrr::map_dfc(setNames, object = list(NA)) %>%
#                          select(!gene_id))
#     } else {
#       # otherwise if gene is present, recover the relevant information from the DE table
#       row <- bind_cols(row,
#                        eval(parse(text = paste0("de_table_", spc))) %>%
#                          filter(grepl(row %>% 
#                                         select(paste0("gene_id_", spc)) %>% 
#                                         sub("\\..*", "", .), 
#                                       gene_id)) %>%
#                          select(!gene_id)
#                        )
#     }
#   }
#   
#   # append current row to output table
#   if(exists("sc_core_4species_merged_table")) {
#     sc_core_4species_merged_table <- bind_rows(sc_core_4species_merged_table, row)
#   } else {
#     sc_core_4species_merged_table <- row
#   }
#   
#   # final message
#   if(nrow == max_rows){
#     cat("\t\t100% of processing done.")
#   }
# }
# 
# write_csv(sc_core_4species_merged_table,
#           file = paste0(out_path, "core_groups_4_species_counts_DE.csv"))


# 5 - extract unique groups for all the species (at least one copy in a single species)
all_species_unique <- tibble()
species <- colnames(ort_freq_table[2:9]) %>% sub("gene_id_", "", .)

for(spc in species){
  spc_unique <- ort_freq_table %>%
    filter(if_all(starts_with("copy_nr") & !paste0("copy_nr_", spc), ~ . == 0))
  all_species_unique <- all_species_unique %>% 
    bind_rows(spc_unique)
}

write_csv(all_species_unique,
          file = paste0(out_path, "unique_groups_all_species.csv"))

# 6 - extract unique groups for H. incana
incana_unique <- ort_freq_table %>%
  filter(if_all(starts_with("copy_nr") & !"copy_nr_h_incana", ~ . == 0))

write_csv(incana_unique,
          file = paste0(out_path, "unique_groups_H_incana.csv"))


# generating table similar as Raúl's
rm(stats_tb)
nr_non_spc_spec_groups <- nrow(ort_freq_table) - nrow(spc_unique)

for(spc in species){
  
  total_genes_spc <- ort_freq_table %>% 
    select(!!rlang::sym(paste0('copy_nr_', spc))) %>% 
    sum()
  
  spc_unique <- ort_freq_table %>%
    filter(if_all(starts_with("copy_nr") & !paste0("copy_nr_", spc), ~ . == 0)) 
  nr_groups_spc_unique <- spc_unique %>% 
    select(!!rlang::sym(paste0('copy_nr_', spc))) %>% 
    nrow()
  nr_genes_spc_unique <- spc_unique %>% 
    select(!!rlang::sym(paste0('copy_nr_', spc))) %>% 
    sum
  pct_genes_spc_unique <- round((nr_genes_spc_unique/total_genes_spc)*100, 2)
  
  nr_groups_spc <- ort_freq_table %>% 
    filter(!!rlang::sym(paste0('copy_nr_', spc)) > 0) %>%  # extract all groups for which thew current species has at least one gene
    nrow()
  
  nr_groups_spc_nonspec <- nr_groups_spc - nr_groups_spc_unique
  pct_groups_spc_nonspec <- round((nr_groups_spc_nonspec/nr_non_spc_spec_groups)*100, 2)
  
  nr_groups_spc_incana <- ort_freq_table %>% 
    filter(!!rlang::sym(paste0('copy_nr_', spc)) > 0 & copy_nr_h_incana > 0) %>% 
    nrow()
  
  col = tibble(a = c(spc, 
                     total_genes_spc,
                     nr_genes_spc_unique,
                     nr_groups_spc_unique,
                     pct_genes_spc_unique,
                     nr_groups_spc_nonspec,
                     nr_groups_spc_incana))
  
  
  if(exists('stats_tb')){
    stats_tb <- stats_tb %>% 
      bind_cols(col)
  } else {
    stats_tb <- col
  }
}

write_csv(stats_tb, 
          file = paste0(out_path, "groups_stats.csv"))
