#!/usr/bin/env python3

import argparse
import numpy as np
import os
import pandas as pd
import random #mostly for testing!
import re
from datetime import datetime
from numpy import nan as NA
#from sys import argv

# functions

def read_ortho_table_dict(table_path):
    """
    create a dictionary to contain all lines of the orthology/frequency table
    """
    lines_dict = {}
    with open(table_path, 'r') as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            elif re.match(r'^\D', line):
                #table_headers = line.split()
                continue
            line_fields = line.split(",")
            #print(line_fields)
            list_line_fields = []
            for field in line_fields[1:5]:
                if re.search(r'\s', field):
                    genes = [x.split('.')[0] for x in field.split(" ")]
                    list_line_fields.append(genes)
                else:
                    gene = field.split('.')[0]
                    list_line_fields.append(gene)
            lines_dict[line_fields[0]] = list_line_fields + line_fields[5:]
    print('succesfully read orthology table')
    return lines_dict

def describe_ortho_table(ortho_dict, csv):
    """
    retrieves some descriptive statistics from the orthology table
    """
    print('Total # of groups: {}\n'.format(len(ortho_dict.keys())))
    dict_res = {'core_all': 0, 'scc_all': 0, 'accessory_all': 0, 'core_4sp': 0,\
                'scc_4sp': 0, 'accessory_4sp':0, 'unique_4sp':0,\
                'excluded_4sp':0}

    #n_core_groups, n_scc_groups, n_core_4sp, n_scc_4sp = 0
    n_unique = [0] * 4

    for values in ortho_dict.values():
        #print(values)
        if all(int(x)>0 for x in values[4:8]):
            dict_res['core_all'] +=1
            if all(int(x)==1 for x in values[4:8]):
                dict_res['scc_all'] +=1 # so scc groups also count as core
        elif values[4:8].count('0') <= 2:
            dict_res['accessory_all'] +=1

        if values[4:8].count('0') == 3:
            for i in range(0, len(n_unique)):
                if int(values[4 + i]) > 0:
                    n_unique[i] +=1

    species = ['A_thaliana', 'B_nigra', 'B_rapa', 'H_incana']
    
    if not csv:
        print('# of core groups (all species):\t{}'.format(dict_res['core_all']))
        print('\t# of scc groups (all species):\t{}'.format(dict_res['scc_all']))
        print('# of accessory groups (all species):\t{}'.format(dict_res['accessory_all']))
        print('# of unique groups (all species):\t{}'.format(sum(n_unique)))

        for i in range(0,len(n_unique)):
            if not csv:
                  print('\t# of unique groups ({}):\t{}'.format(species[i], n_unique[i]))
            else:
                dict_res['unique_'.format(species[i])] = n_unique[i]

    if not csv:
        print()
    else:
        return dict_res
                
def find_gene(gene_id, lines_dict):
    """
    extract the line from the orthology table that contains the gene ID
    """
    for key, value in lines_dict.items():
        if any(gene_id in sublist for sublist in value):
            print('gene ID {} found in group # {}\n'.format(gene_id, key))
            group_id = key
            group_data = value
        else:
            continue
    return group_id, group_data

def find_group(group_id, lines_dict):
    """
    extract the line from the orthology table that contains the group ID
    """
    for key, value in lines_dict.items():
        if group_id == key:
            print('group ID found\n')
            group_id = key
            group_data = value
        else:
            continue
    return group_id, group_data

def read_count_tables(species_list, folder_path):
    """
    automatically read in count tables
    """
    dict_count_tables = {}
    for spc in species_list:
        dict_lines = {}
        file_name = '{}_longest_isoforms_filter_counts_TPM.csv'.\
        format(spc)
        #print('reading {}...'.format(file_name))
        with open('{}{}'.format(folder_path, file_name), 'r') as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                elif re.match(r'^gene_id', line):
                    continue
                line_fields = line.split(',')
                gene_id = line_fields[0].split('.')[0]
                #gene_id = line_fields[0].split('"')[1].split('.')[0]
                # somehow there are "" enclosing the IDs, they give problems
                dict_lines[gene_id] = line_fields[1:]
        dict_count_tables[spc] = dict_lines
    print('succesfully read all count tables')
    return dict_count_tables

def read_DE_tables(species_list, folder_path):
    """
    automatically read in DE tables
    """
    dict_DE_tables = {}
    for spc in species_list:
        dict_lines = {}
        file_name = '{}_DE_results.csv'.\
        format(spc)
        with open('{}{}'.format(folder_path, file_name), 'r') as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                elif line.startswith('""'): # 1st field of 1st line is empty
                    continue
                line_fields = line.split(',')
                dict_lines[line_fields[0].split('"')[1].split('.')[0].split('|')[0]] = line_fields[1:]
        dict_DE_tables[spc] = dict_lines
        print('\t {} DE table has {} rows'.format(spc, len(dict_lines)))
    print('succesfully read all DE tables\n')
    #print(dict_DE_tables['B_rapa']['nbis-gene-145|atpB'])
    return dict_DE_tables

def extract_counts(species_list, dict_count_tables, ortho_key, ortho_values, anno_dict):
    """
    extracts all information for all genes from appropriate count tables
    """
    dict_counts_results = {}
    gene_ids = ortho_values[0:4]
    for i in range(4):
        spc = species_list[i]
        dict_count_tables_spc = dict_count_tables[spc]
        gene_ids_spc = gene_ids[i]
        if len(gene_ids_spc) == 0:
            dict_counts_results['NA.{}.{}'.format(spc, ortho_key)] = [spc] + ['NA']*10
        else:
            if type(gene_ids_spc) is list:
                for ident in gene_ids_spc:
                    if ident in dict_count_tables_spc.keys():
                        counts = dict_count_tables_spc[ident]
                    else:
                        counts = ['NA']*10 # this accounts for filtered-out genes
                
                    if spc == 'A_thaliana':
                        dict_counts_results[ident] = [spc] + [anno_dict[ident]] + counts
                    else:
                        dict_counts_results[ident] = [spc] + [''] + counts
            else:
                ident = gene_ids_spc
                if ident in dict_count_tables_spc.keys():
                    counts = dict_count_tables_spc[ident]
                else:
                    counts = ['NA']*10 # this accounts for filtered-out genes
                if spc == 'A_thaliana':
                    dict_counts_results[ident] = [spc] + [anno_dict[ident]] + counts
                else:
                    dict_counts_results[ident] = [spc] + [''] + counts
                #dict_counts_results[ident_short] = [spc] + counts
    #print(dict_counts_results)        
    return dict_counts_results

def extract_DE(species_list, dict_DE_tables, ortho_key, ortho_values):
    """
    extracts all information for all genes from appropriate DE tables
    """
    dict_DE_results = {}
    gene_ids = ortho_values[0:4]
    #print(dict_DE_tables['B_rapa']['nbis-gene-145|atpB'])
    for i in range(4):
        spc = species_list[i]
        dict_DE_tables_spc = dict_DE_tables[spc]
        gene_ids_spc = gene_ids[i]
        #print(gene_ids_spc)
        if len(gene_ids_spc) == 0:
            dict_DE_results['NA.{}.{}'.format(spc, ortho_key)] = [spc] + ['NA']*6
        else:
            if type(gene_ids_spc) is list:
                for ident in gene_ids_spc:
                    ident_short = ident.split('.')[0]
                    if ident_short in dict_DE_tables_spc.keys():
                        DE_data = dict_DE_tables_spc[ident_short]
                        dict_DE_results[ident_short] = [spc] + DE_data
                    else:
                        dict_DE_results[ident_short] = [spc] + ['NA']*6
            else:
                ident_short = gene_ids_spc.split('.')[0]
                #print(ident_short)
                #print(spc)
                #print(dict_DE_tables_spc.keys())
                #print(dict_DE_tables_spc[ident_short])
                if ident_short in dict_DE_tables_spc.keys():
                    DE_data = dict_DE_tables_spc[ident_short]
                    dict_DE_results[ident_short] = [spc] + DE_data
                else:
                    dict_DE_results[ident_short] = [spc] + ['NA']*6
            
    return dict_DE_results

def count_diff_exp(gene_list, species, dict_DE_results):
    """
    given a list of genes for a species, will check how many are significantly up- or down-regulated
    """
    n_over = 0
    n_under = 0 
    for gene_id in gene_list:
        if not gene_id in DE_dict[species].keys():
            continue
        else:
            de_row = DE_dict[species][gene_id] # this is a list with 6 fields
            if de_row[5] == 'NA':
                continue
            elif float(de_row[5]) < 0.05:
                if float(de_row[1]) > 0 :
                    n_over +=1
                elif float(de_row[1]) < 0:
                    n_under +=1
            else:
                continue

    return [n_over, n_under]

def process_groups_dict(groups_dict, species_list, DE_dict, anno_dict, mode):
    """
    processes extract_groups_type to produce a table with # copies up- and down_regulated per species
    """
    
    if mode == 'groups':
        groups_dict_short = groups_dict
    else:
        groups_dict_short = {}     
        for key,value in groups_dict.items():
            foursp_ids = value[0:4]
            copy_nrs = [int(i) for i in value[4:]] # make sure we're dealing with integers
            short_row = foursp_ids + copy_nrs # output: gene IDs + copies for (At,Br,Bn,Hi)
            groups_dict_short[key] = short_row

    res_dict = {}
    for key, values in groups_dict_short.items():
        over_under_list_all = []
        for i in range(0,4):
            spc = species_list[i]
            genes = values[i]

            if type(genes) is list:
                over_under_list = count_diff_exp(genes, spc, DE_dict)
                over_under_list_all += over_under_list
            else:
                genes = genes.split() # will turn it into a list of 1 string
                over_under_list = count_diff_exp(genes, spc, DE_dict)
                over_under_list_all += over_under_list
                
        at_ids = values[0]
        #print(at_ids)
        #print(anno_dict[at_ids])
        #print(over_under_list_all)

        if not at_ids:
            res_dict[key] = [''] + values + over_under_list_all
        elif type(at_ids) is list:
            list_names = []
            for i in range(0, len(at_ids)):
                list_names += [anno_dict[at_ids[i]]]
            res_dict[key] = [list_names] + values + over_under_list_all
        else:
            res_dict[key] = [anno_dict[at_ids]] + values + over_under_list_all
        #print(values)
        #print(res_dict[key])
    return res_dict

def groups_dict_to_df(res_dict):
    """
    turns dict to df, adds columns allowing to determine overall DE status
    """
    #print(res_dict['110'])
    res_df = pd.DataFrame.from_dict(res_dict, orient = 'index',\
        columns = ['gene_name_At', 'id_At', 'id_Bn', 'id_Br', 'id_Hi', 'copies_At',\
        'copies_Bn', 'copies_Br', 'copies_Hi', 'up_At', 'down_At',\
        'up_Bn', 'down_Bn', 'up_Br', 'down_Br', 'up_Hi', 'down_Hi'])

    res_df['up/copies_At'] = res_df['up_At']/res_df['copies_At']
    res_df['down/copies_At'] = res_df['down_At']/res_df['copies_At']
    res_df['up/copies_Bn'] = res_df['up_Bn']/res_df['copies_Bn']
    res_df['down/copies_Bn'] = res_df['down_Bn']/res_df['copies_Bn']
    res_df['up/copies_Br'] = res_df['up_Br']/res_df['copies_Br']
    res_df['down/copies_Br'] = res_df['down_Br']/res_df['copies_Br']
    res_df['up/copies_Hi'] = res_df['up_Hi']/res_df['copies_Hi']
    res_df['down/copies_Hi'] = res_df['down_Hi']/res_df['copies_Hi']

    res_df['de_copies_ratio_At'] = res_df['up/copies_At'] - res_df['down/copies_At']
    res_df['de_copies_ratio_Bn'] = res_df['up/copies_Bn'] - res_df['down/copies_Bn']
    res_df['de_copies_ratio_Br'] = res_df['up/copies_Br'] - res_df['down/copies_Br']
    res_df['de_copies_ratio_Hi'] = res_df['up/copies_Hi'] - res_df['down/copies_Hi']

    res_df['sum_up/copies'] = res_df['up/copies_At'] + res_df['up/copies_Bn']+\
                              res_df['up/copies_Br'] + res_df['up/copies_Hi']

    res_df['sum_down/copies'] = res_df['down/copies_At'] + \
                                res_df['down/copies_Bn']+ \
                                res_df['down/copies_Br'] +\
                                res_df['down/copies_Hi']

    res_df['is_up_At'] = np.where((res_df['up_At']>0), 1, 0)
    res_df['is_up_Bn'] = np.where((res_df['up_Bn']>0), 1, 0)
    res_df['is_up_Br'] = np.where((res_df['up_Br']>0), 1, 0)
    res_df['is_up_Hi'] = np.where((res_df['up_Hi']>0), 1, 0)

    res_df['is_down_At'] = np.where((res_df['down_At']>0), 1, 0)
    res_df['is_down_Bn'] = np.where((res_df['down_Bn']>0), 1, 0)
    res_df['is_down_Br'] = np.where((res_df['down_Br']>0), 1, 0)
    res_df['is_down_Hi'] = np.where((res_df['down_Hi']>0), 1, 0)
    
    res_df['sum_is_up'] = res_df['is_up_At'] + res_df['is_up_Bn'] + res_df['is_up_Br'] + res_df['is_up_Hi']
    res_df['sum_is_down'] = res_df['is_down_At'] + res_df['is_down_Bn'] + res_df['is_down_Br'] + res_df['is_down_Hi']


    return res_df

def process_single_ID(g_id, type_id, ortho_dict, species_list, counts_dict, \
DE_dict, anno_dict, counts_bool):
    """
    runs all necessary operations to process single gene/group IDs
    """
    print('searching for {}...'.format(g_id))
    # find all gene IDs linked with the gene query
    if type_id == 'gene':
        group_id, group_values = find_gene(g_id, ortho_dict)
    # or find gene IDs associated with a group ID
    elif type_id == 'group': 
        group_id, group_values = find_group(g_id, ortho_dict)

    if counts_bool:
        # get counts data 
        counts_data = extract_counts(species_list, counts_dict, group_id, group_values, anno_dict)
        counts_df = pd.DataFrame.from_dict(counts_data, orient = 'index',\
        columns = ['species', 'gene_name', 'LL1', 'LL2', 'LL3', 'LL4', 'LL5', 'HL1', 'HL2',\
        'HL3', 'HL4', 'HL5'])
        counts_df.index.name = 'gene_id' # handy index naming
        counts_df.insert(0, 'group', group_id) # add group info
        counts_df = counts_df.set_index(['species', counts_df.index]) # multiindex
        counts_df.set_axis(['group', 'gene_name', 'LL1', 'LL2', 'LL3', 'LL4', 'LL5', 'HL1',\
                            'HL2', 'HL3', 'HL4', 'HL5'], axis=1) # renaming columns
    
        # get DE data
        DE_data = extract_DE(species_list, DE_dict, group_id,  group_values)
        DE_df = pd.DataFrame.from_dict(DE_data, orient = 'index',\
        columns = ['species', 'baseMean', 'log2FoldChange', 'lfcSE', 'stat',\
        'pvalue', 'padj'])
        DE_df.index.name = 'gene_id' # handy index naming
        DE_df = DE_df.set_index(['species', DE_df.index]) # multiindex
        DE_df.set_axis(['baseMean', 'log2FoldChange', 'lfcSE', 'stat',\
        'pvalue', 'padj'], axis = 1)


        # merge results into a single table, ready for output
        results_df = pd.concat([counts_df, DE_df], axis=1)
        #results_df = results_df.replace({'NA': np.nan})
        results_df = results_df.drop(labels = ['baseMean', 'lfcSE',\
        'stat', 'pvalue'], axis = 1) # make the table leaner
        results_df = results_df.rename(columns = {'log2FoldChange' : 'log2FC'})
        results_df['log2FC'] = results_df['log2FC'].replace(['NA'], np.nan)
        results_df['padj'] = results_df['padj'].replace(['NA'], np.nan)
        results_df = results_df.astype({'group':'int64', 'log2FC':float,\
                                        'padj':float}, errors = 'ignore')
    
        # add a column with significance symbols
        results_df['signif'] = np.where(results_df['padj']<0.001, '***', \
        np.where(results_df['padj']<0.01, '**', \
        np.where(results_df['padj']<0.05, '*', 'NS')))

        # add a column with up/down regulation symbols
        results_df.insert(12, 'up_down', np.where(results_df['log2FC']>0,\
        '+', np.where(results_df['log2FC'].isnull(), 'NA','-')))
        
    else:
        dict_id = {group_id: group_values}
        dict_group = process_groups_dict(dict_id, species_list, DE_dict,\
                    anno_dict, mode = 'genes')
        #print(dict_group)
        results_df = groups_dict_to_df(dict_group)
        results_df.index.name = 'group_id'

    return results_df        

def process_list_IDs(file_path, type_id, ortho_dict, species_list,\
counts_dict, DE_dict, anno_dict, counts_bool):
    """
    runs process_single_ID over a list of IDs contained in a file
    """
    with open(file_path, 'r') as f:
        group_id_list = []
        for line in f:
            line = line.strip()
            if type_id == 'gene':
                group_id, group_values = find_gene(line, ortho_dict)
                group_id_list.append(group_id)
            elif type_id == 'group':
                group_id_list.append(line)

    group_id_unique = list(set(group_id_list))

    if counts_bool:
        for group_id in group_id_unique:
            if not 'all_df' in locals():
                all_df = process_single_ID(group_id, 'group', ortho_dict,\
                species_list, counts_dict, DE_dict, at_gff_dict, counts_bool)
            else:
                curr_df = process_single_ID(group_id, 'group', ortho_dict,\
                species_list, counts_dict, DE_dict, at_gff_dict, counts_bool)
                all_df = pd.concat((all_df, curr_df), axis = 0)

    else:
        groups_dict = {}
        # this will make a subset dictionary with only the relevant groups
        for group_id in group_id_unique:
            groups_dict[group_id] = ortho_dict[group_id]
            
        all_df = groups_dict_to_df(process_groups_dict(groups_dict,\
                species_list, DE_dict, anno_dict, mode = 'genes'))
        all_df.index.name = 'group_id'

    return all_df

def extract_groups_type(groups_type, dict_orthology):
    """
    filters the orthology table dict to retrieve groups of the specified type
    """
    for key,value in dict_orthology.items():
        #print(key)
        #print(value)
        #indices = [1,2,3,4,8,9,10,11,12,13,14,15]
        foursp_ids = value[:4]
        copy_nrs = [int(i) for i in value[4:]] # make sure we're dealing with integers
        short_row = foursp_ids + copy_nrs # output: gene IDs + copies for (At,Br,Bn,Hi)

        if groups_type == 'core': 
        # NB these groups are defined as 'core' with regards to the (At,Br,Bn,Hi) species' set    
            if 0 in copy_nrs[1:5]:
                continue
            else:
                if not 'core_dict' in locals():
                    core_dict = {key:short_row}
                else:
                    core_dict[key] = short_row

        elif groups_type == 'scc':
        # NB these groups are defined as 'scc' with regards to the (At,Br,Bn,Hi) species' set
            if 0 in copy_nrs:
                continue
            elif any(x > 1 for x in copy_nrs):
                continue
            else:
                if not 'scc_dict' in locals():
                    scc_dict = {key:short_row}
                else:
                    scc_dict[key] = short_row

        elif groups_type == 'accessory':
            if not all(x>=1 for x in copy_nrs):
                if not 'accessory_dict' in locals():
                    accessory_dict = {key:short_row}
                else:
                    accessory_dict[key] = short_row

        elif groups_type == 'unique_at':
        # copy nr. for At is at position 0 of copy_nrs
            if copy_nrs[0] > 0 and all(x == 0 for x in [copy_nrs[i] for i in [1,2,3]]):
                if not 'at_unique_dict' in locals():
                    at_unique_dict = {key:short_row}
                else:
                    at_unique_dict[key] = short_row

        elif groups_type == 'unique_br':
        # copy nr. for Br is at position 2 of copy_nrs
            if copy_nrs[2] > 0 and all(x == 0 for x in [copy_nrs[i] for i in [0,1,3]]):
                if not 'br_unique_dict' in locals():
                    br_unique_dict = {key:short_row}
                else:
                    br_unique_dict[key] = short_row 

        elif groups_type == 'unique_bn':
        # copy nr. for Bn is at position 1 of copy_nrs
            if copy_nrs[1] > 0 and all(x == 0 for x in [copy_nrs[i] for i in [0,2,3]]):
                if not 'bn_unique_dict' in locals():
                    bn_unique_dict = {key:short_row}
                else:
                    bn_unique_dict[key] = short_row 
        elif groups_type == 'unique_hi':
        # copy nr. for Hi is at position 3 of copy_nrs
            if copy_nrs[3] > 0 and all(x == 0 for x in [copy_nrs[i] for i in [0,1,2]]):
                if not 'hi_unique_dict' in locals():
                    hi_unique_dict = {key:short_row}
                else:
                    hi_unique_dict[key] = short_row 
            
    # return appropriate dict
    if groups_type == 'core':
        return core_dict
    elif groups_type == 'scc':
        return scc_dict
    elif groups_type == 'accessory':
        return accessory_dict
    elif groups_type == 'unique_at':
        return at_unique_dict
    elif groups_type == 'unique_br':
        return br_unique_dict
    elif groups_type == 'unique_bn':
        return bn_unique_dict
    elif groups_type == 'unique_hi':
        return hi_unique_dict

def summarise_groups_DE(res_dict, species_list, group_type, csv):
    """
    computes and prints out some descriptive statistics
    """
    all_up = 0
    all_down = 0
    bras_up = 0 # specific to the 3 Brassiceae species
    bras_down = 0
    hi_up = 0 # specific to Hirschfeldia
    hi_down = 0
    all_up_dict = {}
    all_down_dict = {}
    bras_up_dict = {}
    bras_down_dict = {}
    hi_up_dict = {}
    hi_down_dict = {}
    
    total_nr = len(res_dict.keys())
    print('Total # of {} groups in homology table: {}\n'. format(\
        group_type, total_nr))
    
    for key, values in res_dict.items():
        #print()
        #print(key)
        #print(values)
        up_down = values[9:17]
        #print(up_down)
        if all(x > 0 for x in [up_down[i] for i in range(0,8,2)]):
            # this checks if all up-* columns have a value > 0
            all_up +=1
            if csv:
                # will subset res_dict
                all_up_dict[key] = values                
        elif all(x > 0 for x in [up_down[i] for i in range(1,8,2)]):
            # check the down-* columns
            all_down +=1
            if csv:
                all_down_dict[key] = values 
        elif all(x > 0 for x in [up_down[i] for i in range(2,8,2)]) and\
             up_down[0] == 0 :
            # checks the up-[Brassiceae] columns IF the up-At col is zero
            bras_up +=1
            if csv:
                bras_up_dict[key] = values 
        elif all(x > 0 for x in [up_down[i] for i in range(3,8,2)])and\
             up_down[1] == 0 :
            # checks the down-[Brassiceae] columns
            bras_down +=1
            if csv:
                bras_down_dict[key] = values 
        elif all(x == 0 for x in [up_down[i] for i in range(0,6,2)]) and\
             up_down[6] > 0 :
            # checks the up-Hi column IF the up-[Brassiceae] cols are zero
            hi_up +=1
            if csv:
                hi_up_dict[key] = values 
        elif all(x == 0 for x in [up_down[i] for i in range(1,6,2)]) and\
             up_down[7] > 0 :
            # checks the up-Hi column IF the up-[Brassiceae] cols are zero
            hi_down +=1
            if csv:
                hi_down_dict[key] = values
                
    # print the statistics
    print('# of groups w/ genes up-regulated in all species: {}'.format(all_up))
    print('# of groups w/ genes down-regulated in all species: {}'.format(all_down))
    print('# of groups w/ genes exclusively up-regulated in Brassiceae species: {}'.format(bras_up))
    print('# of groups w/ genes exclusively down-regulated in Brassiceae species: {}'.format(bras_down))
    print('# of groups w/ genes exclusively up-regulated in H. incana: {}'.format(hi_up))
    print('# of groups w/ genes exclusively down-regulated in H. incana: {}\n'.format(hi_down))

    if csv:
        # output a list of names for the dicts, and the dicts
        list_names = ['all_up_groups', 'all_down_groups', 'bras_up_groups',\
                      'bras_down_groups', 'hi_up_groups', 'hi_down_groups']
        return list_names, [all_up_dict, all_down_dict, bras_up_dict,\
                            bras_down_dict, hi_up_dict, hi_down_dict]

def import_at_gff(gff_path):
    """
    parse GFF annotation for A. thaliana, return dict of gene IDs and gene names
    """
    dict_id_name = {}
    
    with open(gff_path, 'r') as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            elif line.startswith('#'):
                continue
            else:
                if re.search(r'\smRNA\s', line):
                    id_match = re.search(r'ID=[A-Za-z0-9-:]*', line).group()
                    gene_id = id_match.split('=')[-1]
                    name_match = re.search(r'geneName=.*;[A-Za-z]', line)
                    if name_match:
                        name_match = name_match.group()
                        gene_name = name_match.split(';')[0].split('=')[1]
                    else:
                        name = re.search(r'Name=[A-Za-z0-9]*', line)
                        gene_name = name.group().split('=')[1]
                    dict_id_name[gene_id] = gene_name
                else:
                    continue

    print('succesfully imported A. thaliana annotation\n')
    return dict_id_name


### main ###

if __name__ == '__main__':

    # track running time
    start_time = datetime.now()
    # hardcoded paths
    COUNT_PATH = './data/count_tables/'
    DE_PATH = './data/DE_tables/'
    ORTHO_FILE = './data/renamed_chl_65_orthology_frequency_table.csv'
    SPECIES_LIST = ['A_thaliana', 'B_nigra', 'B_rapa', 'H_incana']
    OUT_PATH = './outputs/'
    AT_GFF_PATH = './data/a_thaliana_Araport11_annotation.gff3'
        # probably the file used for mapping works as well
    
    # create a parser with argparse
    parser = argparse.ArgumentParser(prog = 'analyse_orthology',\
                    description = 'Program can do a number of operations with \
                    orthology, count, and DE tables')

    # will store a gene ID. NB action = 'store' is default
    parser.add_argument('--gene_id', action = 'store', help = 'pass a gene ID \
    and instruct the program to find the orthology group, and all relevant \
    data, connected to the ID') 
    
    # will store the name/path to a text file containing a list of gene IDs
    parser.add_argument('--genes_list', action = 'store', help = 'pass the \
    name of a file containing a list of gene IDs (one per line), and instruct\
    the program to find the orthology group, and all relevant data,\
    connected to the verious IDs') 
    
    # will store the orthology group ID
    parser.add_argument('--group_id', help = 'pass an orthology group ID and\
    instruct the program to retrieve all data connected to its genes')
    
    # will store the name/path to a text file containing a list of group IDs
    parser.add_argument('--groups_list', action = 'store', help = 'pass the \
    name of a file containing a list of group IDs (one per line), and instruct\
    the program to retrieve all data connected to the genes') 
    
    # will store the type of groups weŕe interested in
    parser.add_argument('--groups_type', choices=['core', 'scc', 'accessory',\
    'unique_hi', 'unique_bn', 'unique_br', 'unique_at'], help = \
    'specify a category of orthology groups for which numbers of over- and \
    under-expressed genes will be caclulated and reported') 
    
    # this option allows to redirect output to CSV file instead of screen
    parser.add_argument('--csv', action = 'store_true', help = 'this will \
    make the program output to csv files instead of stdout')

    # this option will report statistics of the orthology table
    parser.add_argument('--ortho_stats', action = 'store_true', help = 'this \
    will make the program output descriptive statistics on the whole orthology\
    table, instead of retrieving DE data')

    # this option allows to get more output detail, namely counts
    parser.add_argument('--counts', action = 'store_true', help = 'this will\
    trigger a different output mode which features per-gene and per-sample\
    read counts' )

    # for testing purposes, currently
    parser.add_argument('--dev', action = 'store_true', help = 'enables \
    development-oriented actions')
    
    # gather passed arguments
    args = parser.parse_args()
    
    print()

    # output folder if --csv is invoked
    if args.csv:
        date_time = datetime.now().strftime('%d-%m-%Y_%H-%M') # timestamp
        if os.path.exists(OUT_PATH):
            print('output will be stored in pre-existing folder {}\n'.format(OUT_PATH))
        else:
            print('creating output folder at {}\n'.format(OUT_PATH))
            os.mkdir(path = OUT_PATH)

    # import ortho table
    ortho_dict = read_ortho_table_dict(ORTHO_FILE)
    #print(ortho_dict['393'])

    # only interested in stats?
    if args.ortho_stats:
        if args.csv:
            describe_ortho_table(ortho_dict, csv = True)
        else:
            describe_ortho_table(ortho_dict, csv = False)
        
    # if not, here comes all the rest
    else:
        
        # gather counts and DE tables
        counts_dict = read_count_tables(SPECIES_LIST, COUNT_PATH)
        DE_dict = read_DE_tables(SPECIES_LIST, DE_PATH)

        #print(DE_dict['A_thaliana']['AT1G01010'])
        #print(DE_dict)
        
        # import the A. thaliana GFF file
        at_gff_dict = import_at_gff(AT_GFF_PATH)

        # run the functions according to passed arguments
        if args.gene_id:
            print('search for gene ID triggered\n')
            GENE_ID = args.gene_id
            if args.counts:
                res_df = process_single_ID(GENE_ID, 'gene', ortho_dict,\
                            SPECIES_LIST, counts_dict, DE_dict, at_gff_dict,\
                            counts_bool = True)
            else:
                res_df = process_single_ID(GENE_ID, 'gene', ortho_dict,\
                            SPECIES_LIST, counts_dict, DE_dict, at_gff_dict,\
                            counts_bool = False)
                
            if args.csv:
                res_df.to_csv('{}{}_{}_results.csv'.format(OUT_PATH, date_time,\
                GENE_ID))
            else:
                print(res_df)

        elif args.genes_list:
            print('search for list of genes triggered\n')
            FILE_PATH = args.genes_list
            if args.counts:
                res_df = process_list_IDs(FILE_PATH, 'gene', ortho_dict,\
                        SPECIES_LIST, counts_dict, DE_dict, at_gff_dict,\
                        counts_bool = True)
            else:
                res_df = process_list_IDs(FILE_PATH, 'gene', ortho_dict,\
                        SPECIES_LIST, counts_dict, DE_dict, at_gff_dict,\
                        counts_bool = False)
            if args.csv:
                res_df.to_csv('{}{}_list-genes_results.csv'.format(OUT_PATH,\
                date_time))
            else:
                print(res_df)

        elif args.group_id:
            print('search for group ID triggered\n')
            GROUP_ID = args.group_id
            if args.counts:
                res_df = process_single_ID(GROUP_ID, 'group', ortho_dict,\
                            SPECIES_LIST, counts_dict, DE_dict, at_gff_dict,\
                            counts_bool = True)
            else:
                res_df = process_single_ID(GROUP_ID, 'group', ortho_dict,\
                            SPECIES_LIST, counts_dict, DE_dict, at_gff_dict,\
                            counts_bool = False)
            if args.csv:
                res_df.to_csv('{}{}_{}_results.csv'.format(OUT_PATH, date_time,\
                GROUP_ID))
            else:
                print(res_df)

        elif args.groups_list:
            print('search for list of groups triggered\n')
            FILE_PATH = args.groups_list
            if args.counts:
                res_df = process_list_IDs(FILE_PATH, 'group', ortho_dict,\
                        SPECIES_LIST, counts_dict, DE_dict, at_gff_dict,\
                        counts_bool = True)
            else:
                res_df = process_list_IDs(FILE_PATH, 'group', ortho_dict,\
                        SPECIES_LIST, counts_dict, DE_dict, at_gff_dict,\
                        counts_bool = False)
            if args.csv:
                res_df.to_csv('{}{}_list-groups_results.csv'.format(OUT_PATH,\
                date_time))
            else:
                print(res_df)
            
        elif args.groups_type:            
            GROUP_TYPE = args.groups_type
            print('search for {} groups triggered\n'.format(GROUP_TYPE))
            groups_dict = extract_groups_type(GROUP_TYPE, ortho_dict)
            res_dict = process_groups_dict(groups_dict, SPECIES_LIST, DE_dict,\
                        at_gff_dict, mode = 'groups')
            res_df = groups_dict_to_df(res_dict)

            print('Results table extract:')
            print(res_df)
            print()

            if args.csv:
                res_df.to_csv('{}{}_{}-groups_results.csv'.format(OUT_PATH,\
                date_time, GROUP_TYPE))
##                list_names, list_dicts = summarise_groups_DE(res_dict,\
##                SPECIES_LIST, GROUP_TYPE, csv = True)
##
##                for i in range(0,len(list_names)):
##                    # outputs the 'broken down' tables
##                    if not len(list_dicts[i]) == 0:
##                        groups_dict_to_df(list_dicts[i]).to_csv('{}{}_{}_{}_results.csv'\
##                        .format(OUT_PATH, date_time, GROUP_TYPE,\
##                                list_names[i]))         
            else:
                summarise_groups_DE(res_dict, SPECIES_LIST, GROUP_TYPE,\
                                    csv = False)
                    
    # report run duration
    end_time = datetime.now()
    print('Run completed in: {}'.format(end_time - start_time))
