#!/usr/bin/env python3

import argparse
import os
import pandas as pd
import re
from datetime import datetime
from analyse_orthology import read_ortho_table_dict, find_gene, find_group,\
     import_at_gff

# functions

def make_GO_dict(path_GO_genes, path_terms_file):
    """
    creates dict[GO:TERM] = [[gene_IDs], [gene_names], description]
    """
    # first, create dict with GO terms as keys and gene IDs, names as values
    dict_gene_go = {}
    go_bp_terms = []
    with open(path_GO_genes, 'r') as f1:
        for line in f1:
            line = line.strip()
            if not line or line.startswith('!'):
                continue
            else:
                fields = re.split(r'\t+', line) # split over tabs!
                # the following will only select biological process terms
                if fields[7] == 'P' or fields[8] == 'P':
                    gene_id = fields[1].upper()
                    gene_name = fields[2]
                    go_term = fields[4]
                    go_bp_terms += [go_term]
                    if gene_id not in dict_gene_go.keys():
                        dict_gene_go[gene_id] = [go_term]
                    else:
                        dict_gene_go[gene_id] += [go_term]
                else:
                    continue

    #print(dict_gene_go['AT2G36830'])
    #print(dict_gene_go.values())

    print('Successfully imported Arabidopsis GO terms.\n\
    Number of GO biological process terms associated with At genes: {}'\
          .format(len(set(go_bp_terms))))
                    
    # now, read the terms file and make dict of biological_process terms
    dict_terms_bp = {}

    #convert all values in dict_gene_go to a set
    set_go_ids = set(go_bp_terms)

    with open(path_terms_file, 'r') as f2:
        lines = f2.readlines()
        # retrieve indices for lines correspodning to 'biological process'
        indices_bp = [i for i,x in enumerate(lines)\
                      if re.search(r'namespace: biological_process', x)]
        # iterate over indices to recover rest of info
        for i in range(0, len(indices_bp)):
            index = indices_bp[i]
            go_id = lines[index-2].strip().split()[1]
            # check if the current GO-ID is in the list of At-linked  IDs
            if go_id in set_go_ids:
                go_name = lines[index-1].strip().split(': ')[1]
                if i < (len(indices_bp)-1):
                    def_index = [x for x in range(index, indices_bp[i+1])\
                                 if re.search(r'^def:', lines[x])].pop(0)
                else:
                    def_index = [x for x in range(index, index+30)\
                                 if re.search(r'^def:', lines[x])].pop(0)

                go_def = lines[def_index].strip().split(': ')[1]
                dict_terms_bp[go_id] = [go_name, go_def]
            # if not, no need to process this and make the dict bigger
            else:
                continue
            
    print('Successfully imported corresponding GO ontology.\n\
    Number of GO biological-process terms: {}\n'\
          .format(len(dict_terms_bp.keys())))

    #print(set_go_ids.difference(dict_terms_bp.keys()))

    # finally, merge the two dicts
    dict_go_complete = {}
    
    for gene_id, go_terms in dict_gene_go.items():
        for go_term in go_terms:
            if go_term in dict_terms_bp.keys():
                if not gene_id in dict_go_complete.keys():
                    dict_go_complete[gene_id] = {go_term : dict_terms_bp[go_term]}
                else:
                    dict_go_complete[gene_id].update({go_term : dict_terms_bp[go_term]})
            else:
                continue
            
    return dict_go_complete

def make_kegg_dict(path_kegg_genes, path_kegg_pathways):
    """
    will make a dict {At_gene : [kegg_path, path_def]}
    """
    kegg_genes_dict = {}
    with open(path_kegg_genes, 'r') as f1:
        for line in f1:
            if not line:
                continue
            line = line.strip()
            fields = re.split(r'\s+', line)
            gene = fields[0].split(':')[1]
            path = fields[1].split(':')[1]
            if gene not in kegg_genes_dict.keys():
                kegg_genes_dict[gene] = [path]
            else:
                kegg_genes_dict[gene] += [path]

    kegg_paths_dict = {}
    with open(path_kegg_pathways, 'r') as f2:
        for line in f2:
            if not line:
                continue
            line = line.strip()
            fields = re.split(r'\t+', line)
            path = fields[0].split(':')[1]
            p_def = fields[1].split(' -')[0]
            kegg_paths_dict[path] = p_def

    kegg_dict = {}
    for key, value in kegg_genes_dict.items():
        list_def = [kegg_paths_dict[x] for x in value]
        kegg_dict[key] = [value, list_def]

    #print(kegg_dict['AT1G09780'])
    return kegg_dict
            

def single_id_go_kegg(g_id, type_id, ortho_dict, go_dict, kegg_dict,\
                      at_gff_dict):
    """
    given an identifier (gene or group), retrieves associated GO terms via At
    """
    print('searching for {}...'.format(g_id))
    # find all gene IDs linked with the gene query
    if type_id == 'gene':
        group_id, group_values = find_gene(g_id, ortho_dict)
    # or find gene IDs associated with a group ID
    elif type_id == 'group': 
        group_id, group_values = find_group(g_id, ortho_dict)

    # retrieve the At gene identifier(s) associated to the gene
    at_ids = group_values[1]
    #print(at_ids)

    if type(at_ids) is str:
        at_ids = [at_ids]

    go_terms = []
    go_names = []
    go_defs = []
    at_genes_no_bp = 0

    kegg_pathways = []
    kegg_defs = []
    at_genes_no_kegg = 0

    at_names = []
    
    for at_id in at_ids:
        # gather the GO terms corresponding to each gene ID
        if at_id in go_dict.keys():
            go_terms += go_dict[at_id].keys()
            go_names += [x[0] for x in go_dict[at_id].values()]
            go_defs +=  [x[1] for x in go_dict[at_id].values()]
        else:
            at_genes_no_bp +=1
        # gather the KEGG terms corresponding to each gene ID
        if at_id in kegg_dict.keys():
            kegg_pathways = kegg_dict[at_id][0]
            kegg_defs = kegg_dict[at_id][1]
        else:
            at_genes_no_kegg +=1
        # gather the At gene names
        at_names += [at_gff_dict[at_id]]


    print('\tgroup {} contains {} Arabidopsis gene(s).\n\
    \tOf those, {} do(es) not have any GO biological process term associated,\n\
    \tand {} do(es) not have any KEGG pathways associated\n'\
          .format(group_id, len(at_ids), at_genes_no_bp, at_genes_no_kegg))

    #print(go_terms)
    gene_ids = group_values[1:5]
    gene_copies = group_values[9:13]
    #line_dict = {group_id: gene_ids + gene_copies + [go_terms] + [go_names] + [go_defs]}
    line_dict = {group_id: gene_ids + gene_copies + [at_names] + [go_terms] + [go_names]\
                 + [kegg_pathways] + [kegg_defs]}

    return line_dict

def multi_id_go(g_ids_list, type_id, ortho_dict, go_dict, kegg_dict,\
                at_gff_dict):
    """
    simply iterates single_id_go over list of IDs
    """
    lines_dict = {}
    for g_id in g_ids_list:
        g_id = g_id.strip()
        line_dict = single_id_go_kegg(g_id, type_id, ortho_dict, go_dict,\
                                      kegg_dict, at_gff_dict)
        lines_dict.update(line_dict)

    return lines_dict

def output_dict(lines_dict, csv, out_path, date_time, mode):
    """
    will either print or save dict
    """
##    dict_df = pd.DataFrame.from_dict(lines_dict, orient = 'index',\
##        columns = ['gene_id_At', 'gene_id_Bn', 'gene_id_Br', 'gene_id_Hi',\
##                   'copies_At', 'copies_Bn', 'copies_Br', 'copies_Hi',\
##                   'GO_terms', 'terms_names', 'terms_defs'])
    dict_df = pd.DataFrame.from_dict(lines_dict, orient = 'index',\
    columns = ['gene_id_At', 'gene_id_Bn', 'gene_id_Br', 'gene_id_Hi',\
               'copies_At', 'copies_Bn', 'copies_Br', 'copies_Hi',\
               'gene_name_At', 'GO_terms', 'terms_names', 'KEGG_pathways',\
               'pathways_names'])
    dict_df.index.name = 'group'
    #dict_df['GO_terms'] = dict_df['GO_terms'].replace(', ', '\n')
    #dict_df['GO_terms'] = dict_df['GO_terms'].apply(clean_alt_list)

    if not csv:
        # will show a super simplified version of the table with GO terms only
        #print(dict_df.iloc[:,8:-1])
        print(dict_df.iloc[:,8:])
    else:
        dict_df.to_csv('{}{}_{}_annotation)results.csv'.format(out_path,\
                date_time, mode))

##def clean_alt_list(list_):
##    list_ = list_.replace(', ', '","')
##    list_ = list_.replace('[', '["')
##    list_ = list_.replace(']', '"]')
##    return list_
##    
    

## main ##

if __name__ == '__main__':

    # track running time
    start_time = datetime.now()

    # hardcoded paths
    ORTHO_FILE = './data/65_orthology_frequency_table_renamed.csv'

    GO_GENES_PATH = './data/ath_GO_tair.gaf'
    GO_TERMS_PATH = './data/go_core_ontology.obo'

    KEGG_GENES_PATH = './data/ath_geneIDs_keggpath.txt'
    KEGG_PATHW_PATH =  './data/ath_keggpath_description.txt'

    AT_GFF_PATH = './data/a_thaliana_Araport11_annotation.longest-isoform_sup_30.gff'
    
    OUT_PATH = './outputs/'

    SPECIES_LIST = ['A_thaliana', 'B_nigra', 'B_rapa', 'H_incana']

    # create a parser with argparse
    parser = argparse.ArgumentParser(prog = 'analyse_orthology',\
                                     description = 'Program can do a number of operations with \
                    orthology, count, and DE tables')

    # will store a gene ID. NB action = 'store' is default
    parser.add_argument('--gene_id', action = 'store', help = 'pass a gene ID \
    and instruct the program to find the orthology group, and all relevant \
    data, connected to the ID') 
    
    # will store the name/path to a text file containing a list of gene IDs
    parser.add_argument('--genes_list', action = 'store', help = 'pass the \
    name of a file containing a list of gene IDs (one per line), and instruct\
    the program to find the orthology group, and all relevant data,\
    connected to the verious IDs') 
    
    # will store the orthology group ID
    parser.add_argument('--group_id', help = 'pass an orthology group ID and\
    instruct the program to retrieve all data connected to its genes')
    
    # will store the name/path to a text file containing a list of group IDs
    parser.add_argument('--groups_list', action = 'store', help = 'pass the \
    name of a file containing a list of group IDs (one per line), and instruct\
    the program to retrieve all data connected to the genes')

    # this option allows to redirect output to CSV file instead of screen
    parser.add_argument('--csv', action = 'store_true', help = 'this will \
    make the program output to csv files instead of stdout')
    
    # for testing purposes, currently
    parser.add_argument('--dev', action = 'store_true', help = 'enables \
    development-oriented actions')
    
    # gather passed arguments
    args = parser.parse_args()

    # nice for stdout
    print()
    
    # output folder if --csv is invoked
    date_time = datetime.now().strftime('%d-%m-%Y_%H-%M') # timestamp
    if args.csv:
        if os.path.exists(OUT_PATH):
            print('output will be stored in pre-existing folder {}\n'.format(OUT_PATH))
        else:
            print('creating output folder at {}\n'.format(OUT_PATH))
            os.mkdir(path = OUT_PATH)

    # import orthology table
    ortho_dict = read_ortho_table_dict(ORTHO_FILE)
    # import GO information
    go_dict = make_GO_dict(GO_GENES_PATH, GO_TERMS_PATH)
    # import KEGG information
    kegg_dict = make_kegg_dict(KEGG_GENES_PATH, KEGG_PATHW_PATH)
    # import Arabidopsis GFF
    at_gff_dict = import_at_gff(AT_GFF_PATH)
    
    
    # process single gene ID
    if args.gene_id:
        GENE_ID = args.gene_id.split('.')[0]
        line_dict = single_id_go_kegg(GENE_ID, 'gene', ortho_dict, go_dict,\
                                      kegg_dict, at_gff_dict)
        output_dict(line_dict, args.csv, OUT_PATH, date_time, 'single_gene')

    # process list of gene IDs    
    elif args.genes_list:
        with open(args.genes_list, 'r') as f:
            GENES_LIST = f.readlines()
        lines_dict = multi_id_go(GENES_LIST, 'gene', ortho_dict, go_dict,\
                                 kegg_dict, at_gff_dict)
        output_dict(lines_dict, args.csv, OUT_PATH, date_time, 'list_genes')

    # process single group ID
    if args.group_id:
        GROUP_ID = args.group_id
        line_dict = single_id_go_kegg(GROUP_ID, 'group', ortho_dict, go_dict,\
                                      kegg_dict, at_gff_dict)
        output_dict(line_dict, args.csv, OUT_PATH, date_time, 'single_group')

    # process list of group IDs    
    elif args.groups_list:
        with open(args.groups_list, 'r') as f:
            GROUPS_LIST = f.readlines()
        lines_dict = multi_id_go(GROUPS_LIST, 'group', ortho_dict, go_dict,\
                                 kegg_dict, at_gff_dict)
        output_dict(lines_dict, args.csv, OUT_PATH, date_time, 'list_groups')        

    # report run duration
    end_time = datetime.now()
    print('Run completed in: {}'.format(end_time - start_time))

