easypackages::packages('topGO', 'KEGGREST', 'org.At.tair.db', 'pheatmap',
                       'RColorBrewer', 'tidyverse')

csv_loc = "C:/MyPrograms/rna_seq_data_analysis/outputs/24-04-2023_14-36_core-groups_results.csv"

data <- read_csv(csv_loc) %>% 
  dplyr::rename(group_id = `...1`)

# retrieve date
ddmm <- paste(rev(str_split_1(as.character(Sys.Date()), '-'))[1:2], collapse = '')

# identifying the within-species mixed responses

mixed_responding_groups <- data %>% filter(if_all(matches('is.*At'), function(x) x==1) |
                                             if_all(matches('is.*Bn'), function(x) x==1) |
                                             if_all(matches('is.*Br'), function(x) x==1) |
                                             if_all(matches('is.*Hi'), function(x) x==1))

nrow(mixed_responding_groups) / nrow(data) *100 # percentage of groups we'll remove for the heatmap of responsive groups

data_hmap <- anti_join(data, mixed_responding_groups, by = 'group_id') %>% 
  filter(`sum_up/copies`!=0 | `sum_down/copies`!=0) %>%  # exclude genes that don't respond
  select(group_id, gene_name_At, de_copies_ratio_At, de_copies_ratio_Bn, 
         de_copies_ratio_Br, de_copies_ratio_Hi)


# plot a heatmap of the whole dataset based on the per-species ratio between de genes and # genes

matrix_hmap <- data_hmap %>% 
  select(!c(group_id, gene_name_At)) %>% 
  as.matrix()

colnames(matrix_hmap) <- c("A. thaliana", "B. nigra", "B. rapa", "H. incana")

hmap <- pheatmap(matrix_hmap, 
                 color = colorRampPalette(rev(brewer.pal(n = 7, name ="RdYlBu")))(100),
                 kmeans_k = NA,
                 scale = "none",
                 fontsize = 14,
                 cluster_rows = TRUE,
                 cluster_cols = TRUE,
                 clustering_distance_rows = "euclidean",
                 clustering_distance_cols = "euclidean",
                 clustering_method = "complete",
                 angle_col = 45)

# now read the DE table to import gene names and p-values

de_at_loc <- "../../rna_seq_data_analysis/data/DE_tables/A_thaliana_DE_results.csv"

de_data <-  read_csv(de_at_loc) %>% 
  filter(!is.na(padj)) %>% 
  separate(`...1`, c('gene_id', 'other'), sep = '.Ara') %>% 
  select(gene_id, padj)

genes_at <- de_data %>% pull(gene_id)
#padjs_at <- de_data %>% pull(padj)

# initialize a set of vectors to collect statistics

list_names = character()
list_counts = numeric()

list_names <- append(list_names, 'total number of groups')
list_counts <- append(list_counts, nrow(data))

## - iterate over the various categories of groups and perform a GO and/or KEGG enrichment - ##

# this will serve to name the stats list
group_names <- c('non-responsive groups', 'up-regulated, not all species', 
                 'up-regulated, all species', 'up-regulated, Brassiceae',
                 'up-regulated, incana-only', 'down-regulated, not all species',
                 'down-regulated, all species', 'down-regulated, Brassiceae',
                 'down-regulated, incana-only', 'up- or down-regulated between species',
                 'up- or down-regulated within species')

# this will serve to iterate, select & so on
group_categories <- c('non_responsive_genes', 'up_DE_genes', 
                      'up_DE_genes_all_spp', 'up_DE_genes_brassiceae',
                      'up_DE_genes_hi', 'down_DE_genes',
                      'down_DE_genes_all_spp', 'down_DE_genes_brassiceae',
                      'down_DE_genes_hi', 'updown_interspecies_DE_genes',
                      'updown_intraspecies_DE_genes')

# want to perform GO enrichment?
GO_sel <- F
# want to perform KEGG enrichment?
KEGG_sel <- F

# iterate
for (i in 1:length(group_categories)) {
  cat = group_categories[i]
  
  print(paste0('processing ', group_names[i], '...'))
  
  # select the groups that do not respond to treatment 
  if (cat == 'non_responsive_genes'){
    data_sel <- data %>% 
      filter(`sum_up/copies`==0 & `sum_down/copies`==0)
  }
  # select the groups that do respond positively to treatment, but not in all species
  else if(cat == 'up_DE_genes'){
    # NB in some cases these Arabidopsis genes do not respond per se,
    # but are orthologs of genes that do respond
    data_sel <- data %>% 
      filter(sum_is_up < 4 & sum_is_up >0 & sum_is_down == 0)
  }
  # select the groups that respond positively to treatment in all species
  else if (cat == 'up_DE_genes_all_spp'){
    data_sel <- data %>% 
      filter(sum_is_up ==4 & sum_is_down == 0)
  }
  # select the groups that respond positively to treatment in the 3 Brassiceae
  else if(cat == 'up_DE_genes_brassiceae'){
    data_sel <- data %>% 
      filter(is_up_Hi ==1 & is_up_At ==0 & is_up_Bn ==1 & is_up_Br ==1 & sum_is_down == 0)
  } 
  # select the groups that respond positively to treatment in Hirschfeldia alone
  else if(cat == 'up_DE_genes_hi'){
    data_sel <- data %>% 
      filter(is_up_Hi ==1 & is_up_At ==0 & is_up_Bn ==0 & is_up_Br ==0 & sum_is_down == 0)
  }
  # select the groups that do respond negatively to treatment, but not in all species
  else if(cat == 'down_DE_genes'){
    # NB in some cases these Arabidopsis genes do not respond per se,
    # but are orthologs of genes that do respond
    data_sel <- data %>% 
      filter(sum_is_down < 4 & sum_is_down >0 & sum_is_up == 0)  
  } 
  # select the groups that respond negatively to treatment in all species
  else if(cat == 'down_DE_genes_all_spp'){
    data_sel <- data %>% 
      filter(sum_is_down ==4 & sum_is_up == 0)
  }
  # select the groups that respond negatively to treatment in the 3 Brassiceae
  else if(cat == 'down_DE_genes_brassiceae'){
    data_sel <- data %>% 
      filter(is_down_Hi ==1 & is_down_At ==0 & is_down_Bn ==1 & is_down_Br ==1 & sum_is_up == 0)
  }
  # select the groups that respond negatively to treatment in Hirschfeldia only
  else if(cat == 'down_DE_genes_hi'){
    data_sel <- data %>% 
      filter(is_down_Hi ==1 & is_down_At ==0 & is_down_Bn ==0 & is_down_Br ==0 & sum_is_up == 0)
  }
  # select the genes that have a mixed response (up- + down-) between species
  else if(cat == 'updown_interspecies_DE_genes'){
    data_updown <- data %>% 
      filter(if_any(matches('is_up'), function(x) x==1) &
               if_any(matches('is_down'), function(x) x==1))
    
    data_sel <- anti_join(data_updown, mixed_responding_groups, by = "group_id")
  }
  # select the genes that have a mixed response (up- + down-) within at least one species
  else if(cat == 'updown_intraspecies_DE_genes'){
    data_sel <- mixed_responding_groups
  }
  
  list_names <- append(list_names, group_names[i])
  list_counts <- append(list_counts, nrow(data_sel))
    
  genes_sel <- data_sel %>% 
    pull(id_At)
    
  all_genes_sel <- character()
    
  for (element in genes_sel) {
    genes <- str_split_1(element, ' ') 
    all_genes_sel <- append(all_genes_sel, genes)
    }

  #write_delim(tibble(all_genes_sel),
              paste0('./outputs/', cat, '_list_', ddmm,'.txt'),
              delim = '', col_names = FALSE)
      
  gene_list_sel <- ifelse((genes_at %in% all_genes_sel), 1, 0)
  names(gene_list_sel) <- genes_at
  
  if (GO_sel){
    cat('\trunning GO enrichment analysis')
    GOdata_sel <- new("topGOdata",
                     ontology = "BP",
                     allGenes = gene_list_sel,
                     geneSelectionFun = function(x) (x==1),
                     annot = annFUN.org, mapping = "org.At.tair.db")
    
    results_sel <- runTest(GOdata_sel, algorithm = "classic", statistic = "fisher")
    
    tab_GO_sel <- GenTable(GOdata_sel, raw.p.value = results_sel, 
                       topNodes = length(results_sel@score), numChar = 120) %>% 
      filter(`raw.p.value` < 0.05) %>% 
      mutate(type = cat)
    
    if (i == 1){
      tab_GO_all <- tab_GO_sel
    } else {
      tab_GO_all <- bind_rows(tab_GO_all, tab_GO_sel)
    }
  }
  
  if (KEGG_sel){
    cat('\trunning KEGG enrichment analysis\n')
    KEGG_results_sel <- clusterProfiler::enrichKEGG(all_genes_sel,
                            organism = 'ath',
                            #universe = genes_at,
                            keyType = 'kegg')
    
    tab_KEGG_sel <- KEGG_results_sel@result %>% 
      filter(p.adjust < 0.05) %>% 
      mutate(type = cat)
    
    if (i == 1){
      tab_KEGG_all <- tab_KEGG_sel
    } else {
      tab_KEGG_all <- bind_rows(tab_KEGG_all, tab_KEGG_sel)
    }
  }
  cat('done\n')
}

# export tables

write_csv(tab_GO_all %>% filter(Significant >= 5),
          paste0('./outputs/GO_analysis_results_', ddmm,'.csv'))
write_csv(tab_KEGG_all, paste0('./outputs/KEGG_analysis_results_', ddmm,'.csv'))

stats_tab <- tibble('type' = list_names, 'n_groups' = list_counts)
write_csv(stats_tab, paste0('./outputs/core_groups_stats_', ddmm, '.csv'))

count_terms <- tab_GO_all %>% 
  count(`GO.ID`) %>% 
  arrange(desc(n)) %>%  
  filter(n > 1)

nrow(count_terms) # number of terms occurring more than once across various sets of groups

# calculating total # of groups containing responses for individual species
data_single_up <- data %>% 
  filter((is_up_Hi ==1 & is_up_At ==0 & is_up_Bn ==0 & is_up_Br ==0 & sum_is_down == 0) |
           (is_up_Hi ==0 & is_up_At ==1 & is_up_Bn ==0 & is_up_Br ==0 & sum_is_down == 0) |
           (is_up_Hi ==0 & is_up_At ==0 & is_up_Bn ==1 & is_up_Br ==0 & sum_is_down == 0) |
           (is_up_Hi ==0 & is_up_At ==0 & is_up_Bn ==0 & is_up_Br ==1 & sum_is_down == 0))

data_single_up %>% select(starts_with('is_up'), starts_with('is_down')) %>% 
  pheatmap(cluster_cols = F,
           show_rownames = F)

data_single_down <- data %>% 
  filter((is_down_Hi ==1 & is_down_At ==0 & is_down_Bn ==0 & is_down_Br ==0 & sum_is_up == 0) |
           (is_down_Hi ==0 & is_down_At ==1 & is_down_Bn ==0 & is_down_Br ==0 & sum_is_up == 0) |
           (is_down_Hi ==0 & is_down_At ==0 & is_down_Bn ==1 & is_down_Br ==0 & sum_is_up == 0) |
           (is_down_Hi ==0 & is_down_At ==0 & is_down_Bn ==0 & is_down_Br ==1 & sum_is_up == 0))

data_single_down %>% select(starts_with('is_up'), starts_with('is_down')) %>% 
  pheatmap(cluster_cols = F,
           show_rownames = F)